/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// MD2 Header
typedef struct
{
	int	ident;				// magic number. must be equal to "IPD2"
	int	version;			// md2 version. must be equal to 8

	int	skinwidth;			// width of the texture
	int	skinheight;			// height of the texture
	int	framesize;			// size of one frame in bytes

	int	num_skins;			// number of textures
	int	num_xyz;			// number of vertices
	int	num_st;				// number of texture coordinates
	int	num_tris;			// number of triangles
	int	num_glcmds;			// number of opengl commands
	int	num_frames;			// total number of frames

	int	ofs_skins;			// offset to skin names (64 bytes each)
	int	ofs_st;				// offset to s-t texture coordinates
	int	ofs_tris;			// offset to triangles
	int	ofs_frames;			// offset to frame data
	int	ofs_glcmds;			// offset to opengl commands
	int	ofs_end;			// offset to end of file

} MD2Header;

// Vertex
typedef struct
{
    unsigned char	v[3];				// compressed vertex (x, y, z) coordinates
    unsigned char	lightnormalindex;	// index to a normal vector for the lighting

} Vertex;

// Texture Coordinates
typedef struct
{
    int s;
    int t;

} Coordinate;

// Frame
typedef struct
{
    float scale[3];			// scale values
    float translate[3];		// translation vector
    char name[16];			// frame name
    Vertex verts[1];		// first vertex of this frame

} Frame;

// triangle
typedef struct
{
    short index_xyz[3];		// indexes to triangle's vertices
    short index_st[3];		// indexes to vertices' texture coorinates

} Triangle;

// animation
typedef struct
{
    int first_frame;		// first frame of the animation
    int last_frame;			// number of frames
    int fps;				// number of frames per second

} Animation;

// animation state
typedef struct
{
    int startframe;		// first frame
    int endframe;		// last frame
    int fps;			// frame per second for this animation

    float curr_time;	// current time
    float old_time;		// old time
    float interpol;		// percent of interpolation

    int type;			// animation type

    int curr_frame;		// current frame
    int next_frame;		// next frame

} AnimState;

// ------------------ The actual model --------------------

class MD2Model : public Linkable {

	private:

		int num_frames;         // number of frames
		int num_xyz;            // number of vertices
		int num_glcmds;         // number of opengl commands

		vec3_t *m_vertices;        // vertex array
		int *m_glcmds;          // opengl command array
		int *m_lightnormals;    // normal index array

		unsigned int m_texid;            // texture id
		animState_t m_anim;             // animation
		float m_scale;            // scale value

	void Animate(float time);
	void ProcessLighting(void);
	void Interpolate(vec3_t *vertlist);
	void RenderFrame(void);

	public:

	    static vec3_t anorms[NUMVERTEXNORMALS];
	    static float anorms_dots[SHADEDOT_QUANT][256];

    	static anim_t animlist[21]; // animation list

    CMD2Model(void);
    ~CMD2Model(void);

    // functions
    bool LoadModel(const char *filename);
    bool LoadSkin(const char *filename);

    void DrawModel(float time);
    void DrawFrame(int frame);

    void SetAnim(int type);
    void ScaleModel(float s) {m_scale = s;}

};
