/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "battle.h"

void initBattle()
{
	debug(("initBattle()\n"));

	loadGameLocaleData("data/locale/battle", true);
	loadGameLocaleData("data/locale/loadSave", true);
	loadGameLocaleData("data/locale/options", true);
	loadGameLocaleData("data/locale/weapons", true);
	loadGameLocaleData("data/locale/items", true);
	loadGameLocaleData("data/locale/status", true);
	loadGameLocaleData("data/locale/controls", true);
	loadGameLocaleData("data/locale/cheats", true);

	// load this mission's locale data
	loadGameLocaleData(mission->localeFile.getText(), true);

	loadWidgets("data/widgets/inGameOptions");
	loadWidgets("data/widgets/generalOptions");
	loadWidgets("data/widgets/controlOptions");;
	loadWidgets("data/widgets/status");
	loadWidgets("data/widgets/loadSave");
	loadWidgets("data/widgets/gameover");
	loadWidgets("data/widgets/cheats");

	loadAllBattleResources();
	loadMapAudio();

	// --------- init battledata ------------------

	game->setMissionStatus(MISSION_INPROGRESS);

	GLfloat ambient[]= {1.0f, 1.0f, 1.0f, 0.0f};
	GLfloat diffuse[]= {1.0f, 1.0f, 1.0f, 0.0f};
	GLfloat specular[]= {1.0f, 1.0f, 1.0f, 0.0f};
	GLfloat position[]= {1.0f, 1.0f, 1.0f, 0.0f};

	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
	glLightfv(GL_LIGHT0, GL_POSITION, position);
	glEnable(GL_LIGHT0);
	
	GLColor fogColor = mission->fogColor;
	
	glClearColor(fogColor.color[0], fogColor.color[1], fogColor.color[2], fogColor.color[3]);

	glFogi(GL_FOG_MODE, GL_LINEAR);
	glHint(GL_FOG_HINT, GL_NICEST);
	
	glFogfv(GL_FOG_COLOR, fogColor.color);
	glFogf(GL_FOG_DENSITY, mission->fogDensity);
	glFogf(GL_FOG_START, mission->fogMin);
	glFogf(GL_FOG_END, mission->fogMax);
	glEnable(GL_FOG);
	
	debug(("initBattle() :: Constructing BSP Bounding Box Grid\n"));
	
	bbManager->construct(bsp->minCords, bsp->maxCords);

	debug(("initBattle() :: Done\n"));
}

void prepareBattleScene()
{
	static Vector idealPosition;
	
	graphics->setMode(MODE_3D);

	if (game->missionStatus != MISSION_PLAYER_OUT)
	{
		if ((camera->targetCamera == NULL) || (!game->doingCutscene))
		{
			camera->position = player->position;
			camera->velocity.set(game->cameraDir.x, -game->cameraDir.y, game->cameraHeight);
			camera->target = player->position;
			
			checkEntityToWorld(camera);
			camera->position = tracer->outputEnd;
			camera->contentType = tracer->outputContent;
			
			float distance = Math::getDistance(player->position.x, player->position.y, tracer->outputEnd.x, tracer->outputEnd.y);
			
			if (distance < 171)
			{
				float adjust = game->cameraHeight / 171.0;
				
				camera->velocity.set(0, 0, -(game->cameraHeight - (adjust * distance)));
				
				checkEntityToWorld(camera);
				
				camera->position = tracer->outputEnd;
			}
			
			/*
			* We need to compensate for the fact that the angle the camera is at might be completely
			* different to the angle the player is moving in. ie - Ensure pushing forward always moves
			* the player up the screen and not in another direction.
			*/
			float dx = player->position.x - camera->position.x;
			float dy = player->position.y - camera->position.y;
		 
			game->cameraAdjustedRadian = atan2(dy, dx);
			game->cameraAdjustedRadian = -90 + toDegrees(game->cameraAdjustedRadian);
			game->cameraAdjustedRadian = toRadians(game->cameraAdjustedRadian);
		}
		else
		{
			camera->position = camera->targetCamera->position;
			camera->target = camera->targetCamera->destination;
		}
	}
	
	camera->use();

	glColor3f(1.0, 1.0, 1.0);
	
	glFogf(GL_FOG_START, mission->fogMin);
	glFogf(GL_FOG_END, mission->fogMax);
}

void doPause()
{
	debug(("Player = %s\n", player->position.toString()));
	
	game->resetControl(CTRL_PAUSE);

	audio->pause();

	mouse->setBusy(false);

	while (true)
	{
		graphics->updateScreen();

		engine->getInput();

		if (game->isControl(CTRL_PAUSE))
		{
			break;
		}
	}

	engine->resetTimeDifference();

	audio->resume();

	game->resetControl(CTRL_PAUSE);
	
	mouse->setBusy(true);
}

void drawGameObjects()
{
	// this is nasty hack... We assume this is the temporary title screen... I'll fix it later :)
	if (!game->inGame)
	{
		graphics->setMode(MODE_2D);
		glColor3f(1.0, 1.0, 1.0);
		drawTitle();
		return;
	}
	
	prepareBattleScene();
	
	textureSpecManager->resetTextureAnimation();
	
	drawSkyDome(camera->position, 500, textureManager->getTexture("SkySphere"));
	
	// draw all solid objects
	
	drawBSP(false);
	
	Reference *ref;
	
	for (ref = (Reference*)entityManager->drawList.getFirstElement() ; ref != NULL ; ref = (Reference*)ref->next)
	{
		self = (Entity*)ref->object;
		
		if (self->draw != NULL)
		{
			self->draw(false);
		}
		else
		{
			printf("ERROR: draw() for entity %d (%s) is NULL\n", self->entityType, self->getName());
			exit(1);
		}
	}
	
	drawBSP(true);
	
	// draw all solid transparent objects
	
	for (ref = (Reference*)entityManager->drawList.getFirstElement() ; ref != NULL ; ref = (Reference*)ref->next)
	{
		self = (Entity*)ref->object;
		
		if (self->draw != NULL)
		{
			self->draw(true);
		}
	}
	
	#if DEV
	drawTriggers();
	#endif
	
	// finally generate all the shadows...
	
	for (ref = (Reference*)entityManager->drawList.getFirstElement() ; ref != NULL ; ref = (Reference*)ref->next)
	{
		self = (Entity*)ref->object;
		
		generateShadow();
	}
	
	drawParticles();
	
	drawShadow();
	
	drawSelectedUnit();
	
	drawGrapplingHook();
	
	GLColor c;
	
	if (game->whiteOutTime > 0)
	{
		graphics->setMode(MODE_2D);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		c = GLColor::white;
		c.color[3] = game->whiteOutTime;
		game->whiteOutTime -= (0.01 * engine->getTimeDifference(TD_LOGIC));
		graphics->drawRect(0, 0, graphics->screen->w, graphics->screen->h, c, true);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		graphics->setMode(MODE_3D);
	}
		
	if (camera->contentType & CONTENTS_WATER)
	{
		c = GLColor::skyBlue;
	}
	else if (camera->contentType & CONTENTS_SLIME)
	{
		c = GLColor::green;
	}
	else if (camera->contentType & CONTENTS_LAVA)
	{
		c = GLColor::red;
	}
	else
	{
		return;
	}
	
	c.color[3] *= 0.25;
	graphics->setMode(MODE_2D);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	graphics->drawRect(0, 0, graphics->screen->w, graphics->screen->h, c, true);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	graphics->setMode(MODE_3D);
	
	camera->contentType = CONTENTS_SOLID;
}

void doGameObjects()
{
	entityManager->drawList.clear();
	
	doItems();
	
	doBullets();
	
	doEnemies();
	doBlobs();
	
	renderBSP();
	
	doStructures();
	
	doDecoration();
	
	doTriggers();
	doFeatures();
	doParticles();
	doWeather();
	
	doEmitters();
	
	doTraps();
}

void doBattleCamera()
{
	if (engine->keyState[SDLK_F2])
	{
		if ((game->canSaveGame) || (game->cheatSave))
		{
			saveGame();
			engine->resetTimeDifference();
		}

		engine->keyState[SDLK_F2] = 0;
	}

	if (engine->keyState[SDLK_F5])
	{
		if ((game->canSaveGame) || (game->cheatSave))
		{
			saveGameData(false);
			engine->resetTimeDifference();
		}

		engine->keyState[SDLK_F5] = 0;
	}
	
	if (engine->keyState[SDLK_F9])
	{
		engine->keyState[SDLK_F9] = 0;
		engine->showFPS = !engine->showFPS;
		engine->showPosition = !engine->showPosition;
	}
	
	if (engine->keyState[SDLK_F11])
	{
		engine->keyState[SDLK_F11] = 0;
		graphics->wireframe = !graphics->wireframe;
	}
	
	if (engine->keyState[SDLK_F12])
	{
		engine->keyState[SDLK_F12] = 0;
		graphics->lightMaps = !graphics->lightMaps;
	}
	
	if (engine->keyState[SDLK_ESCAPE])
	{
		engine->keyState[SDLK_ESCAPE] = 0;
		doInGameOptions();
	}
}

int mainBattleLoop()
{
	debug(("mainBattleLoop()\n"));
	
	game->inGame = true;

	initBattle();
	
	if (game->changingLevel)
	{
		game->changingLevel = false;
		saveGameData(true);
	}
	
	game->retrying = false;

	graphics->clearScreen();

	uiManager->onlyShowGroup("standardPanel");

	int rtn = SECTION_BATTLE;

	mouse->setBusy(true);
	engine->slowLogic = false;
	
	//player->position.set(-1728.116333, 3318.197998, 67.031250);
	
	debug(("Player start: %s\n", player->position.toString()));
	
	Vector explosionPosition;
	game->timeUpExplosionTime = Math::rrand(10, 75);
	
	float bannerTimeOut = 525;
	
	graphics->setFont(FONT_DIGITAL);
	String banner = localeManager->translate("MISSION_NAME");
	banner.toLowerCase();
	Texture *missionBanner = graphics->getGLString(banner.getText());
	textureManager->removeTexture("MissionBanner");
	textureManager->addTexture("MissionBanner", missionBanner);
	
	float autoInfoTimer = 50;

	game->changeCameraRotation(-game->cameraAdjustedRadian);

	graphics->resetFPSCount();
	
	float spawnTime = Math::rrand(mission->minSpawnTime, mission->maxSpawnTime);
	float spawnInterval = Math::rrand(50, 150);
	int numberToSpawn = Math::rrand(min(mission->enemyLevel, mission->spawnSize), mission->spawnSize);

	camera->setOwner(player);
	camera->contentType = CONTENTS_SOLID;
	camera->targetCamera = NULL;
	
	engine->resetTimeDifference();
	
	audio->playSoundLoop(SND_WEATHER, CH_ENVIRONMENT);
	audio->playMusic();
	
	debug(("mainBattleLoop() - Begin\n"));

	while (rtn == SECTION_BATTLE)
	{
		engine->doTimeDifference();
		graphics->updateScreen();
		engine->getInput();
		
		if (game->isControl(CTRL_PAUSE))
		{
			game->resetControl(CTRL_PAUSE);
			doPause();
		}
		
		graphics->clearScreen();
		
		//SDL_Delay(50);
		
		if (game->missionStatus != MISSION_INPROGRESS)
		{
			game->missionOverTimer -= engine->getTimeDifference(TD_LOGIC);
		}
		
		if (mission->spawnSize > 0)
		{
			if ((spawnTime -= engine->getTimeDifference(TD_LOGIC)) <= 0)
			{
				if ((spawnInterval -= engine->getTimeDifference(TD_LOGIC)) <= 0)
				{
					if (isOkayToSpawnEnemies())
					{
						spawnRandomEnemy();
						spawnInterval = Math::rrand(10, 75);
						numberToSpawn--;
					}
					else
					{
						numberToSpawn = 0;
					}
					
					if (numberToSpawn == 0)
					{
						numberToSpawn = Math::rrand(min(mission->enemyLevel, mission->spawnSize), mission->spawnSize);
						spawnTime = Math::rrand(mission->minSpawnTime, mission->maxSpawnTime);
					}
				}
			}
		}

		doGameObjects();
		drawGameObjects();

		doPlayer();

		game->incTime(0.01 * engine->getTimeDifference(TD_LOGIC));
		game->decTimeLimit(0.01 * engine->getTimeDifference(TD_LOGIC));
		game->doComboTimer(engine->getTimeDifference(TD_LOGIC));

		graphics->setMode(MODE_2D);
		
		drawInfoPanel();

		#if DEV
		graphics->drawFPSCount(5);
		#endif
		
		if (bannerTimeOut > 0)
		{
			bannerTimeOut -= engine->getTimeDifference(TD_LOGIC);
			
			if (bannerTimeOut > 400)
			{
				glColor4f(1.0, 1.0, 1.0, (500 - bannerTimeOut) / 100);
				graphics->blit(missionBanner, graphics->screen->w / 2, graphics->screen->h - 180, true);
			}
			else if ((bannerTimeOut > 100) && (bannerTimeOut <= 400))
			{
				glColor3f(1.0, 1.0, 1.0);
				graphics->blit(missionBanner, graphics->screen->w / 2, graphics->screen->h - 180, true);
			}
			else if (bannerTimeOut > 0)
			{
				glColor4f(1.0, 1.0, 1.0, bannerTimeOut / 100);
				graphics->blit(missionBanner, graphics->screen->w / 2, graphics->screen->h - 180, true);
			}
		}
		
		if (mission->autoInfo != "")
		{
			autoInfoTimer -= engine->getTimeDifference(TD_LOGIC);
			
			if (autoInfoTimer <= 0)
			{
				showMessageDialog(localeManager->translate(mission->autoInfo.getText()), "SYS_OK");
				mission->autoInfo = "";
			}
		}

		if ((game->missionStatus != MISSION_INPROGRESS) && (game->missionOverTimer <= 0))
		{
			switch (game->missionStatus)
			{
				case MISSION_COMPLETE:
					saveGameData(true);
					rtn = SECTION_CHANGE_LEVEL;
					break;
				
				case MISSION_PLAYER_OUT:
					graphics->clearScreen();
					graphics->delay(500, false);
					resetPlayer();
					resetEnemyThinkTimes();
					break;
				
				case MISSION_FAILED:
					rtn = doGameOver();
					break;
					
				case MISSION_LOADING:
					rtn = SECTION_LOAD;
					break;
					
				case MISSION_QUITTING:
					rtn = SECTION_TITLE;
					break;
			}
		}
		
		if ((mission->timeLimit != 0) && (game->totalRemaining <= 1))
		{
			game->timeUpExplosionTime -= engine->getTimeDifference(TD_LOGIC);
			
			if (game->timeUpExplosionTime <= 0)
			{
				explosionPosition = player->position;
				explosionPosition.x += Math::rrand(-150, 150);
				explosionPosition.y += Math::rrand(-150, 150);
				explosionPosition.z += Math::rrand(-10, 10);
				doExplosion(explosionPosition, 15, player);
				game->timeUpExplosionTime = Math::rrand(10, 35);
				if ((rand() % 3) == 0)
				{
					game->whiteOutTime = Math::rrand(0, 150);
					game->whiteOutTime /= 100;
				}
			}
		}

		doBattleCamera();
		
		if (camera->targetCamera != NULL)
		{
			doInGameCutscene();
		}
	}

	if (game->missionStatus != MISSION_FAILED)
	{
		audio->stopMusic();
		graphics->clearScreen();
		graphics->updateScreen();
	}

	return rtn;
}
