/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "bullets.h"

void assignBulletActionType(Bullet *bullet);
void addClusterBomblet(Vector position, Entity *owner);

void drawBullet(bool transparent)
{
	if (transparent)
	{
		return;
	}
	
	Bullet *bullet = (Bullet*)self;
	
	if (!(bullet->flags & EF_EXPLODES))
	{
		glEnable(GL_TEXTURE_2D);
		
		glPushMatrix();
		{
			glTranslatef(bullet->position.x, bullet->position.y, bullet->position.z);
			
			if (((Unit*)bullet->owner)->entityType == ET_BLOB)
			{
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, textureManager->getTexture("Bullet")->data);
				glColor4fv(bullet->color.color);
			}
			else
			{
				glDisable(GL_TEXTURE_2D);
				glColor4fv(GLColor::getRandomColor().color);
			}
			
			glRotatef(bullet->rotation.x, 0.0, 0.0, 1.0);
			glRotatef(bullet->rotation.y, 0.0, 1.0, 0.0);
			glRotatef(bullet->rotation.z, 1.0, 0.0, 0.0);
			
			glCallList(modelManager->getSimpleModel("Bullet")->data);
		}
		glPopMatrix();
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
		
		glPushMatrix();
		{
			glColor4fv(GLColor::green.color);
			
			glTranslatef(bullet->position.x, bullet->position.y, bullet->position.z);
			
			gluSphere(graphics->quadric, 1, 4, 8);
		}
		glPopMatrix();
	}
}

void doBulletUnitHit(Unit *unit, int damage, int damageType)
{
	Bullet *bullet = (Bullet*)self;
	
	bool wasAlive = (unit->health > 0);
	
	float shield = unit->shield;
	float maxShield = unit->maxShield;
	
	unit->applyDamage(damage, damageType);
	
	if (unit == player)
	{
		game->alterStat("TOTAL_DAMAGE_TAKEN", damage);
	}
	
	// shield was destroyed in previous hit!
	if ((unit->shield == -2.5) && (shield > 0))
	{
		addShieldDestroyedParticles(unit);
		unit->shieldHit = 0;
	}
	
	// shield was completely destroyed! (player only)
	if (unit == player)
	{
		if ((unit->maxShield == 0) && (maxShield > 0))
		{
			addShieldCompletelyDestroyedParticles(unit);
			unit->shieldHit = 0;
			game->setInfoMessage(INFO_BAD, localeManager->translate("INFO_SHIELD_DESTROYED"));
		}
	}
	
	if (bullet->owner != NULL)
	{
		if ((wasAlive) && (unit->health <= 0) && (unit->entityType == ET_BIOMECH))
		{
			switch (bullet->owner->entityType)
			{
				case ET_BLOB:
					if (bullet->owner == player)
					{
						game->alterStat("ENEMIES_KILLED", 1);
					}
					else
					{
						game->alterStat("ENEMIES_KILLED_TEEKA", 1);
					}
					break;
				
				case ET_BIOMECH:
					game->alterStat("ENEMIES_KILLED_FRIENDLY", 1);
					break;
			}
		}
	}
				
	game->alterStat("BULLETS_HIT", 1);
		
	if (unit == player)
	{
		game->alterStat("TOTAL_DAMAGE_TAKEN", damage);
	}
	
	// first check for combo hits
	if ((unit->health <= 0) && (unit->flags & EF_DYING))
	{
		if (!(unit->flags & EF_STATIC))
		{
			throwUnit(unit, bullet->velocity);
			Math::removeBit(&unit->flags, EF_DYING);
		}
		
		if ((bullet->owner == player) || (bullet->owner->definition->type == BLOB_TEEKA))
		{
			game->comboTimer = 150;
			game->comboHits++;
		}
		
		return;
	}
	
	if ((unit->maxShield == 0) || (unit->shield <= 0))
	{
		if (!(unit->flags & EF_EXPLODES))
		{
			addBloodParticles(bullet);
			audio->playSound(SND_HIT, CH_ANY, camera->getSoundDistance(bullet->position));
		}
		else
		{
			addSparkParticle(bullet->position);
			audio->playSound(SND_CLANG, CH_ANY, camera->getSoundDistance(bullet->position));
		}
	}
	else
	{
		if (unit->shield > 0)
		{
			addShieldHitParticles(bullet->position);
			audio->playSound(SND_ELECTRIC_SPARK4, CH_ANY, camera->getSoundDistance(bullet->position));
		}
	}
	
	if (unit->health <= 0)
	{
		// force of impact - should be overridden in die() if needed
		if (!(unit->flags & EF_STATIC))
		{
			throwUnit(unit, bullet->velocity);
		}
		
		if (bullet->owner == player)
		{
			game->comboTimer = 150;
			game->comboHits++;
		}
	}
	else
	{
		if (!(unit->flags & EF_STATIC))
		{
			unit->velocity.x += (bullet->velocity.x * 0.05);
			unit->velocity.y += (bullet->velocity.y * 0.05);
		}
		unit->reload += 5;
	}
	
	if ((unit != player) && (unit->target == NULL) && (!(unit->flags & EF_STATIC)))
	{
		unit->target = (Unit*)bullet->owner;
		unit->thinkTime = 0;
	}
}

void doBulletDamage(Entity *other, int damage)
{
	if (self->health == 0)
	{
		return;
	}
	
	if ((other == self->owner) && (self->owner != NULL))
	{
		return;
	}
	
	Bullet *bullet = (Bullet*)self;
	WeaponDef *def = (WeaponDef*)bullet->definition;
	
	bullet->health = 0;
	
	if (other != NULL)
	{
		if (other->entityType == ET_BULLET)
		{
			return;
		}
		
		if ((other->entityType != ET_BIOMECH) && (bullet->owner == player))
		{
			game->alterStat("BULLETS_MISSED", 1);
		}
		
		Unit *unit;
		
		switch (other->entityType)
		{
			case ET_BLOB:
			case ET_BIOMECH:
				unit = (Unit*)other;
				doBulletUnitHit(unit, damage, def->damageType);
				if (unit->maxHealth > 5)
				{
					if (unit->totalCurrentDamage >= (unit->maxHealth / 5))
					{
						if ((rand() % 5) == 0)
						{
							throwUnit(unit, bullet->velocity);
						}
					}
				}
				break;
				
			case ET_WINDOW:
				other->health -= damage;
				addSparkParticle(bullet->position);
				break;
				
			case ET_SUPPLY_CRATE:
				other->health -= damage;
				addSparkParticle(bullet->position);
				break;
				
			case ET_BARRIER:
				addSparkParticle(bullet->position);
				audio->playSound(SND_ELECTRIC_SPARK4, CH_ANY, camera->getSoundDistance(bullet->position));
				break;
				
			case ET_STRUCTURE:
				if (!(other->flags & EF_IMMORTAL))
				{
					other->health -= damage;
					addSparkParticle(bullet->position);
				}
				else
				{
					addBulletHitParticles(bullet);
					audio->playSound(Math::rrand(SND_RICO1, SND_RICO2), CH_ANY, camera->getSoundDistance(bullet->position));
				}
				break;
				
			default:
				addBulletHitParticles(bullet);
				audio->playSound(Math::rrand(SND_RICO1, SND_RICO2), CH_ANY, camera->getSoundDistance(bullet->position));
				break;
		}
	}
	else
	{
		addBulletHitParticles(bullet);
		audio->playSound(Math::rrand(SND_RICO1, SND_RICO2), CH_ANY, camera->getSoundDistance(bullet->position));
		
		if (bullet->owner == player)
		{
			game->alterStat("BULLETS_MISSED", 1);
		}
	}
}

void clusterLauncherHit(Entity *other)
{
	if (self->owner == other)
	{
		return;
	}
	
	self->health = 0;
	
	for (int i = 0 ; i < 10 ; i++)
	{
		addClusterBomblet(self->position, other);
	}
}

void rocketLauncherHit(Entity *other)
{
	if (self->owner == other)
	{
		return;
	}
	
	if (other != NULL)
	{
		if (other->health < 50)
		{
			// kill it instantly
			other->health = 0;
		}
		else if (other->entityType == ET_BIOMECH)
		{
			Unit *unit = (Unit*)other;
			unit->shield = unit->maxShield = 0;
			unit->health -= 25;
			
			if (!(other->flags & EF_STATIC))
			{
				unit->velocity = self->velocity;
			}
		}
	}
	
	self->health = 0;
}

void explosiveBulletHit(Entity *other)
{
	if ((other == self->owner) && (self->owner != NULL))
	{
		return;
	}
	
	self->health = 0;
}

void bulletHitSingleDamage(Entity *other)
{
	doBulletDamage(other, 1);
}

void bulletHitDamage(Entity *other)
{
	Bullet *bullet = (Bullet*)self;
	
	doBulletDamage(other, ((WeaponDef*)bullet->definition)->damage);
}

void iceShotHit(Entity *other)
{
	self->health = 0;
	
	if (other == NULL)
	{
		return;
	}
	
	if ((other->entityType != ET_BLOB) && (other->entityType != ET_BIOMECH))
	{
		return;
	}
	
	if (other->flags & EF_FROZEN)
	{
		return;
	}
	
	Unit *unit = (Unit*)other;
	
	unit->mental -= 15;
	
	if (unit->mental <= 0)
	{
		Math::addBit(&unit->flags, EF_FROZEN);
		Math::addBit(&unit->flags, EF_PUSHABLE);
		
		unit->thinkTime = 0;
		unit->action = generalUnitFrozen;
		
		float frozenSize = max(unit->boundingBox.maxs.x, unit->boundingBox.maxs.y);
		frozenSize = max(unit->boundingBox.maxs.z, frozenSize);
		frozenSize++;
		
		unit->boundingBox.mins.set(-frozenSize, -frozenSize, -frozenSize);
		unit->boundingBox.maxs.set(frozenSize, frozenSize, frozenSize);
	}
}

void desimilatorHit(Entity *other)
{
	self->health = 0;
	
	if (other == NULL)
	{
		return;
	}
	
	if ((other == NULL) || (other->entityType != ET_BIOMECH) || (other->flags & EF_FROZEN) || (other->health < 1))
	{
		return;
	}
	
	Unit *unit = (Unit*)other;
	
	if ((!(unit->flags & EF_NOTARGET)) && (!game->cheatDesimilate))
	{
		return;
	}
	
	unit->mental -= 15;
	
	if (unit->mental <= 0)
	{
		// create a new MIA blob based on the enemy's data
		Properties props;
		
		props.setProperty("name", unit->getName());
		props.setProperty("position", unit->position.toString());
		props.setProperty("classname", "BLOB_MIA");
		props.setProperty("health", unit->health);
		props.setProperty("maxHealth", unit->maxHealth);
		props.setProperty("shield", unit->shield);
		props.setProperty("maxShield", unit->maxShield);
		props.setProperty("helpless", 100);
		props.setProperty("velocity", "0 0 0.25");
		props.setProperty("flags", EF_BOUNCES);
		
		addBlob(&props);
		
		// now remove the old enemy...
		unit->health = 0;
		Math::addBit(&unit->flags, EF_DEAD);
		
		addDesimilatedParticles(unit);
		
		fireTriggers("ANY_ENEMY", TRIGGER_TYPE_ENEMY_DESIMILATED);
	}
}

void grenadeHit(Entity *other)
{
	if (other == NULL)
	{
		if (self->realVelocity.z >= EPSILON)
		{
			audio->playSound(SND_METAL_BOUNCE, CH_ANY, camera->getSoundDistance(self->position));
		}
		
		return;
	}
	
	if (self->owner == other)
	{
		return;
	}
	
	switch (other->entityType)
	{
		case ET_DOOR:
		case ET_SWITCH:
		case ET_PUSHBLOCK:
		case ET_LIFT:
		case ET_AUTOLIFT:
			audio->playSound(SND_METAL_BOUNCE, CH_ANY, camera->getSoundDistance(self->position));
			break;
			
		case ET_STRUCTURE:
			if (other->flags & EF_IMMORTAL)
			{
				audio->playSound(SND_METAL_BOUNCE, CH_ANY, camera->getSoundDistance(self->position));
			}
			else
			{
				self->health = 0;
			}
			break;
			
		default:
			self->health = 0;
			break;
	}
}

void bulletDieExplode()
{
	Bullet *bullet = (Bullet*)self;
	
	doExplosion(bullet->position, ((WeaponDef*)bullet->definition)->damage, bullet->owner);
}

void miniMineStick(Entity *entity)
{
	self->velocity.set(0, 0, 0);
	
	Math::addBit(&self->flags, EF_WEIGHTLESS);
}

void addNormalBullet(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	destination -= origin;
	destination.normalize();
	destination.x *= BULLET_MOVE_SPEED;
	destination.y *= BULLET_MOVE_SPEED;
	destination.z *= BULLET_MOVE_SPEED;
	
	Bullet *bullet = entityManager->spawnBullet(definition->type);
	bullet->setOwner(owner);
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 200;
	bullet->rotation = owner->rotation;
	
	assignBulletActionType(bullet);
	
	if (owner->entityType == ET_BLOB)
	{
		bullet->color = GLColor::white;
	}
	else
	{
		bullet->color = GLColor::purple;
	}
}

void addIceShot(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	destination -= origin;
	destination.normalize();
	destination.x *= BULLET_MOVE_SPEED;
	destination.y *= BULLET_MOVE_SPEED;
	destination.z *= BULLET_MOVE_SPEED;
	
	Bullet *bullet = entityManager->spawnBullet(definition->type);
	bullet->setOwner(owner);
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 50;
	bullet->rotation = owner->rotation;
	
	assignBulletActionType(bullet);
	
	if (owner->entityType == ET_BLOB)
	{
		bullet->color = GLColor::white;
	}
	else
	{
		bullet->color = GLColor::purple;
	}
}

void addDesimilatorBeam(Unit *owner, Vector origin, Vector destination)
{
	destination -= origin;
	destination.normalize();
	destination.x *= BULLET_MOVE_SPEED;
	destination.y *= BULLET_MOVE_SPEED;
	destination.z *= BULLET_MOVE_SPEED;
	
	Bullet *bullet = entityManager->spawnBullet(WEP_DESIMILATOR);
	bullet->setOwner(owner);
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 200;
	bullet->rotation = owner->rotation;
	
	assignBulletActionType(bullet);
	
	bullet->color = GLColor::white;
}

void addPellet(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	float x = Math::rrand(-200, 200);
	float y = Math::rrand(-200, 200);
	float z = Math::rrand(-5, 5);
	
	x /= 50;
	y /= 50;
	z /= 25;
	
	destination.x += x;
	destination.y += y;
	destination.z += z;
	
	destination -= origin;
	destination.normalize();
	destination.x *= BULLET_MOVE_SPEED;
	destination.y *= BULLET_MOVE_SPEED;
	destination.z *= BULLET_MOVE_SPEED;
	
	Bullet *bullet = entityManager->spawnBullet(definition->type);
	bullet->setOwner(owner);
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 200;
	bullet->rotation = owner->rotation;
	
	assignBulletActionType(bullet);
	
	if (owner->entityType == ET_BLOB)
	{
		bullet->color = GLColor::white;
	}
	else
	{
		bullet->color = GLColor::purple;
	}
}

void fireWeaponBeam(Unit *owner, WeaponDef *weaponDef, Vector origin, Vector destination)
{
}

void throwGrenade(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	Bullet *bullet = entityManager->spawnBullet(definition->type);
	bullet->setOwner(owner);
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->rotation = owner->rotation;
	
	bullet->velocity = destination;
	bullet->velocity -= origin;
	
	float steps;
	steps = (fabs(origin.x - destination.x));
	steps = max(steps, (fabs(origin.y - destination.y)));
	
	float adjust = 0.5;
	
	if (owner != player)
	{
		float extra = rand() % 25;
		extra /= 100;
		adjust += extra;
	}
	
	bullet->velocity.x /= steps;
	bullet->velocity.y /= steps;
	bullet->velocity.z = (steps / 2) * (GRAVITY * adjust);
	
	bullet->health = 100 + steps;
	
	assignBulletActionType(bullet);
	
	owner->reload = definition->reloadTime;
}

void addRocket(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	destination -= origin;
	destination.normalize();
	destination.x *= BULLET_MOVE_SPEED;
	destination.y *= BULLET_MOVE_SPEED;
	destination.z *= BULLET_MOVE_SPEED;
	
	Bullet *bullet = entityManager->spawnBullet(definition->type);
	bullet->setOwner(owner);
	bullet->definition = definition;
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 500;
	bullet->rotation = owner->rotation;
	
	assignBulletActionType(bullet);
}

void addGrenadeLauncherShot(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	Bullet *bullet;
	
	bullet = entityManager->spawnBullet(definition->type);
	bullet->setOwner(owner);
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 9999;
	bullet->rotation = owner->rotation;
	
	bullet->velocity = destination;
	bullet->velocity -= origin;
	
	float steps;
	steps = (fabs(origin.x - destination.x));
	steps = max(steps, (fabs(origin.y - destination.y)));
	
	bullet->velocity.x /= steps;
	bullet->velocity.y /= steps;
	bullet->velocity.z = (steps / 2) * (GRAVITY * 0.55);
	
	assignBulletActionType(bullet);
	
	owner->reload = definition->reloadTime;
}

void addClusterBomblet(Vector position, Entity *owner)
{
	Bullet *bullet = entityManager->spawnBullet(WEP_BOMBLET);
	bullet->setOwner(owner);
	bullet->position = position;
	bullet->health = 1000;
	
	bullet->velocity.x = Math::rrand(-50, 50);
	bullet->velocity.y = Math::rrand(-50, 50);
	bullet->velocity.z = Math::rrand(50, 100);
	
	float velocity = Math::rrand(100, 150);
	velocity *= 0.0001;
	
	bullet->velocity = bullet->velocity * velocity;
	
	assignBulletActionType(bullet);
}

void addMiniMine(Unit *owner, Vector position, Vector destination)
{
	Bullet *bullet = entityManager->spawnBullet(WEP_MINI_MINE);
	bullet->setOwner(owner);
	bullet->position = position;
	bullet->health = 850;
	
	bullet->velocity = bullet->velocity * BULLET_MOVE_SPEED;
	
	assignBulletActionType(bullet);
}

void assignBulletActionType(Bullet *bullet)
{
	bullet->draw = &drawBullet;
	
	switch (bullet->definition->type)
	{
		case WEP_SHOTGUN:
			bullet->touch = bulletHitSingleDamage;
			break;
		
		case WEP_GRENADE:
		case WEP_MINI_GRENADES:
			bullet->touch = grenadeHit;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_GRENADE_LAUNCHER:
			bullet->touch = explosiveBulletHit;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_CLUSTER_LAUNCHER:
			bullet->touch = clusterLauncherHit;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_ROCKET_LAUNCHER:
			bullet->touch = rocketLauncherHit;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_MINI_ROCKET:
			bullet->touch = explosiveBulletHit;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_BOMBLET:
			bullet->touch = explosiveBulletHit;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_MINI_MINE:
			bullet->touch = miniMineStick;
			bullet->die = bulletDieExplode;
			break;
			
		case WEP_ICE_GUN:
			bullet->touch = iceShotHit;
			break;
			
		case WEP_DESIMILATOR:
			bullet->touch = desimilatorHit;
			break;
		
		default:
			bullet->touch = bulletHitDamage;
			break;
	}
}

void addBullet(Properties *props)
{
	Bullet *bullet = entityManager->spawnBullet(WEP_PISTOL);
	
	int type = props->getInt("definitionType", 0);
	
	bullet->definition = &entityManager->weaponDef[type];
	
	bullet->load(props);
	
	assignBulletActionType(bullet);
}

/*
 * Will show blood if required
*/
void doBulletHitEffect(Bullet *bullet, Unit *unit)
{
	if (!(unit->flags & EF_EXPLODES))
	{
		addBloodParticles(bullet);
	}
	else
	{
		addSparkParticle(bullet->position);
	}
	
	game->alterStat("BULLETS_HIT", 1);
}

void doBullets()
{
	static WeaponDef *def;
	
	glDisable(GL_BLEND);
	
	for (Bullet *bullet = (Bullet*)entityManager->bulletList.getFirstElement() ; bullet != NULL ; bullet = (Bullet*)bullet->next)
	{
		self = bullet;
		
		bbManager->removeBox(bullet);
		
		if (bullet->health <= 0)
		{
			if (bullet->die != NULL)
			{
				bullet->die();
			}
			
			bullet = (Bullet*)bullet->previous;
			entityManager->bulletList.remove(bullet->next);
			continue;
		}
		
		def = (WeaponDef*)bullet->definition;
		
		if (bullet->owner != NULL)
		{
			bullet->owner->referenced = true;
		}
		
		if (withinViewableRange(bullet))
		{
			entityManager->addEntityToDrawList(bullet);
		}
		
		bullet->applyGravity(engine->getTimeDifference(TD_LOGIC));
		
		moveEntity(bullet);
		
		bullet->health -= engine->getTimeDifference(TD_LOGIC);
		
		if ((tracer->outputContent & CONTENTS_WATER) && (!(bullet->contentType & CONTENTS_WATER)))
		{
			audio->playSound(SND_BULLET_HIT_WATER, CH_EFFECT1, camera->getSoundDistance(bullet->position));
		}
		
		bullet->contentType = tracer->outputContent;
		
		if (bullet->contentType & CONTENTS_WATER)
		{
			addBubbleParticle(bullet->position);
		}
		
		if (def->trailType != TRAIL_NONE)
		{
			addTrailParticle(bullet);
		}
		
		bbManager->addBox(bullet);
	}
}
