/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "cutscene.h"

void addCutsceneCamera(Properties *props)
{
	char *name = props->getString("name", "unknown");
	
	CutsceneCamera *camera = entityManager->spawnCutsceneCamera(name);
	
	camera->load(props);
}

void activateInGameCutscene(char *name, float delay)
{
	debug(("activateInGameCutscene - %s, %.2f\n", name, delay));
	
	for (CutsceneCamera *cutsceneCamera = (CutsceneCamera*)entityManager->cameraList.getFirstElement() ; cutsceneCamera != NULL ; cutsceneCamera = (CutsceneCamera*)cutsceneCamera->next)
	{
		if (cutsceneCamera->name == name)
		{
			if (cutsceneCamera->custom != 0)
			{
				continue;
			}
			
			camera->targetCamera = cutsceneCamera;
			camera->targetCamera->custom = delay;
			return;
		}
	}
}

CutsceneCamera *getCutsceneCamera(char *name)
{
	for (CutsceneCamera *cutsceneCamera = (CutsceneCamera*)entityManager->cameraList.getFirstElement() ; cutsceneCamera != NULL ; cutsceneCamera = (CutsceneCamera*)cutsceneCamera->next)
	{
		if (cutsceneCamera->name == name)
		{
			if (cutsceneCamera->custom != 0)
			{
				continue;
			}
			
			return cutsceneCamera;
		}
	}
	
	return NULL;
}

void haltObjectsForCutscene()
{
	entityManager->bulletList.clear();
	
	for (Unit *unit = (Unit*)entityManager->blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		unit->velocity.x = unit->velocity.y = 0;
	}
	
	for (Unit *unit = (Unit*)entityManager->enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		unit->velocity.x = unit->velocity.y = 0;
	}
}

void doInGameCutscene()
{
	engine->clearInput();
	engine->flushInput();
	
	bool doneTriggers = false;
	
	CutsceneCamera *targetCamera = camera->targetCamera;
	
	targetCamera->velocity = targetCamera->destination;
	targetCamera->velocity -= targetCamera->position;
	targetCamera->velocity.normalize();
	targetCamera->velocity = targetCamera->velocity * targetCamera->speed;
	
	while (targetCamera->custom > -targetCamera->timeout)
	{
		engine->doTimeDifference();
		
		graphics->updateScreen();
		engine->getInput();
		
		graphics->clearScreen();
		
		doGameObjects();
		drawGameObjects();
		haltObjectsForCutscene();
		
		targetCamera->custom -= engine->getTimeDifference(TD_LOGIC);
		
		targetCamera->position += (targetCamera->velocity * engine->getTimeDifference(TD_LOGIC));
		
		if (targetCamera->custom <= 0)
		{
			game->doingCutscene = true;
			
			if ((targetCamera->custom <= -25) && (!doneTriggers))
			{
				if (targetCamera->deferredSwitch != NULL)
				{
					handleStructureStateChange(targetCamera->deferredSwitch->getName(), targetCamera->deferredSwitch->active);
					toggleLaserTrap(targetCamera->deferredSwitch->getName());
					fireTriggers(targetCamera->deferredSwitch->getName(), TRIGGER_TYPE_ANY);
				}
				
				fireTriggers(targetCamera->getName(), TRIGGER_TYPE_ANY);
				
				if (mission->allObjectiveComplete())
				{
					fireTriggers("MISSION_COMPLETE", TRIGGER_TYPE_MISSION_COMPLETE);
				}
				
				doneTriggers = true;
			}
		}
		
		if (engine->userAccepts())
		{
			targetCamera->custom = -targetCamera->timeout;
		}
	}
	
	game->doingCutscene = false;
	
	camera->targetCamera = NULL;
	
	engine->clearInput();
	engine->flushInput();
	
	resetEnemyThinkTimes();
}
