/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "eyeDroids.h"

// ###################### General Stuff ############################

void generalDroidDie()
{
	((Unit*)self)->health = Math::rrand(-25, 0);
	self->velocity.z = 0;
	
	((Unit*)self)->helpless = 200;

	audio->playSound(Math::rrand(SND_DROID_DEATH1, SND_DROID_DEATH3), CH_ANY, camera->getSoundDistance(self->position));

	generalUnitDropWeaponAndItems();

	Math::removeBit(&self->flags, EF_WEIGHTLESS);
}

void generalDroidTouch(Entity *other)
{
	if (((Unit*)self)->helpless == 0)
	{
		return;
	}
	
	self->health = -100;
}

// ###################### Pistol Droids ############################

void pistolDroidAttack()
{
	//debug(("pistolDroidAttack()\n"));

	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}

	self->thinkTime = 0;
	self->action = &pistolDroidAttack;

	attackTarget(unit);

	if (--self->custom == 0)
	{
		self->action = &bioMechSearch;
		self->thinkTime = Math::rrand(10, 50);
	}
}

void pistolDroidPreAttack()
{
	//debug(("pistolDroidPreAttack()\n"));

	// move in a random direction before firing...
	self->velocity.x = Math::rrand(-4, 4);
	self->velocity.y = Math::rrand(-4, 4);
	self->velocity.z = Math::rrand(-4, 4);

	self->velocity.x *= 0.01;
	self->velocity.y *= 0.01;
	self->velocity.z *= 0.01;

	self->custom = Math::rrand(1, 2);
	self->thinkTime = 0;
	self->action = &pistolDroidAttack;
}

void pistolDroidInit(Unit *unit)
{
	unit->walk = &bioMechSearch;
	unit->touch = &generalDroidTouch;
	unit->attack = &pistolDroidPreAttack;
	unit->die = &generalDroidDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}

// ###################### Machine Gun Droids ############################

void machineGunDroidAttack()
{
	//debug(("machineGunDroidAttack()\n"));

	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}
	
	self->thinkTime = 0;
	self->action = &machineGunDroidAttack;

	attackTarget(unit);

	if (--self->custom == 0)
	{
		self->action = &bioMechSearch;
		self->thinkTime = Math::rrand(10, 50);
	}
}

void machineGunDroidPreAttack()
{
	//debug(("machineGunDroidPreAttack()\n"));

	self->custom = Math::rrand(4, 12);
	self->thinkTime = 0;
	self->action = &machineGunDroidAttack;
}

void machineGunDroidInit(Unit *unit)
{
	unit->walk = &bioMechSearch;
	unit->touch = &generalDroidTouch;
	unit->attack = &machineGunDroidPreAttack;
	unit->die = &generalDroidDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}

// ###################### Shotgun Droids ############################

void shotGunDroidAttack()
{
	//debug(("shotGunBlobAttack()\n"));

	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}

	self->thinkTime = 0;
	self->action = &shotGunDroidAttack;

	attackTarget(unit);
	
	unit->reload = Math::rrand(25, 50);

	if (--self->custom == 0)
	{
		self->action = &bioMechSearch;
		self->thinkTime = Math::rrand(5, 25);
	}
}

void shotGunDroidPreAttack()
{
	self->custom = Math::rrand(1, 3);
	self->thinkTime = 0;
	self->action = &shotGunDroidAttack;
}

void shotGunDroidInit(Unit *unit)
{
	unit->walk = &bioMechSearch;
	unit->touch = &generalDroidTouch;
	unit->attack = &shotGunDroidPreAttack;
	unit->die = &generalDroidDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}

// ###################### Grenade Droids ############################

void grenadeGunDroidUziAttack()
{
	Unit *unit = (Unit*)self;
	
	unit->currentWeapon = &entityManager->weaponDef[WEP_UZI];

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}
	
	attackTarget(unit);
	
	if (--self->custom == 0)
	{
		self->action = &bioMechSearch;
		self->thinkTime = Math::rrand(5, 25);
	}
}

void grenadeGunDroidGrenadeAttack()
{
	Unit *unit = (Unit*)self;
	
	unit->currentWeapon = &entityManager->weaponDef[WEP_MINI_GRENADES];

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}
	
	attackTarget(unit);
	
	if (--self->custom == 0)
	{
		self->action = &bioMechSearch;
		self->thinkTime = Math::rrand(200, 300);
	}
}

void grenadeGunDroidPreAttack()
{
	if ((rand() % 4) == 0)
	{
		self->custom = Math::rrand(6, 10);
		self->action = grenadeGunDroidUziAttack;
	}
	else
	{
		
		self->custom = Math::rrand(3, 4);
		self->action = grenadeGunDroidGrenadeAttack;
	}
	
	self->thinkTime = 0;
}

void grenadeDroidInit(Unit *unit)
{
	unit->walk = &bioMechSearch;
	unit->touch = &generalDroidTouch;
	unit->attack = &grenadeGunDroidPreAttack;
	unit->die = &generalDroidDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}
