/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "localeLoader.h"

void loadLocaleIndex()
{
	Properties *properties = &LocaleManager::getInstance()->localeIndex;
	
	if (!engine->loadData("data/locale/index"))
	{
		printf("ERROR: Couldn't load Locale Index File (data/locale/index)");
		exit(1);
	}

	char key[1024];
	char value[1024];

	char *token = strtok((char*)engine->dataBuffer, "\n");

	bool done = false;

	while (!done)
	{
		// ignore blank lines
		if (strcmp(token, "") == 0)
		{
			token = strtok(NULL, "\n");
			continue;
		}

		// looks like a comment.
		if (token[0] == '/')
		{
			token = strtok(NULL, "\n");
			continue;
		}

		if (strstr(token, "="))
		{
			if (strstr(token, ";"))
			{
				sscanf(token, "%[^=]%*c%[^;]", key, value);
				properties->setProperty(key, value);
			}
			else
			{
				printf("WARNING: Locale line '%s' does not end with a semicolumn! Skipping.\n", token);
			}
		}
		else if (strcmp(token, "@END@") == 0)
		{
			done = true;
		}

		token = strtok(NULL, "\n");
	}
}

Properties *loadLocaleData(char *filename)
{
	char path[PATH_MAX];
	sprintf(path, "%s", filename);

	debug(("Loading locale data from '%s'\n", path));

	if (!engine->loadData(path))
	{
		printf("WARNING - Could not load '%s'!!\n", path);
		return NULL;
	}
	
	Properties *properties = new Properties();

	char key[1024];
	char value[1024];
	char localeSymbol[10];
	
	char *token = strtok((char*)engine->dataBuffer, "\n");

	bool done = false;
	
	while (!done)
	{
		// ignore blank lines
		if ((strcmp(token, "") == 0) || (strlen(token) == 0))
		{
			token = strtok(NULL, "\n");
			continue;
		}
		
		// looks like a comment.
		if (token[0] == '/')
		{
			token = strtok(NULL, "\n");
			continue;
		}
		
		if (strcmp(token, "@END@") == 0)
		{
			done = true;
		}
		else if (!strstr(token, ";"))
		{
			token = strtok(NULL, "\n");
			continue;
		}
		else
		{
			sscanf(token, "%s %s %[^;]", key, localeSymbol, value);
			
			if (strlen(localeSymbol) != 2)
			{
				printf("WARNING: Locale symbol '%s' in file '%s' must be exactly 2 characters!!\n", localeSymbol, filename);
				token = strtok(NULL, "\n");
				continue;
			}
			
			// Looks like the default string...
			if (strcmp(localeSymbol, "en") == 0)
			{
				properties->setProperty(key, value);
			}
			else
			{
				if (strcmp(localeSymbol, localeManager->getLanguage()) == 0)
				{
					printf("Added %s = %s\n", key, value);
					
					properties->setProperty(key, value);
				}
			}
		}

		token = strtok(NULL, "\n");
	}

	return properties;
}

void loadGameLocaleData(char *filename, bool append)
{
	Properties *properties = loadLocaleData(filename);

	if (properties == NULL)
	{
		return;
		Graphics::getInstance()->showErrorAndExit(ERR_FILE, filename);
	}

	if (!append)
	{
		debug(("Setting Game Properties to Locale Manager...\n"));
		LocaleManager::getInstance()->setLocaleData(properties);
	}
	else
	{
		debug(("Appending Game Properties to Locale Manager...\n"));
		LocaleManager::getInstance()->appendLocaleData(properties);
		delete properties;
	}
}

void loadSystemLocaleData(char *filename)
{
	Properties *properties = loadLocaleData(filename);

	if (properties == NULL)
	{
		return;
		Graphics::getInstance()->showErrorAndExit(ERR_FILE, filename);
	}

	debug(("Setting System Properties to Locale Manager...\n"));
	LocaleManager::getInstance()->setSystemLocaleData(properties);
}
