/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "modelLoader.h"

void createDisplayListFromSimpleModel(char *filename, SimpleModel *model)
{
	if (!engine->loadData(filename))
	{
		graphics->showErrorAndExit(ERR_FILE, filename);
	}
	
	char *token = strtok((char*)engine->dataBuffer, "\n");
	
	int numberOfPoints = atoi(token);
	
	if ((numberOfPoints % 3) != 0)
	{
		printf("ERROR - loadSimpleModel() - Incorrect format! Must be divisible by 3!!\n");
		exit(1);
	}
	
	float x[3], y[3], z[3], u[3], v[3];
	
	model->data = glGenLists(1);
	
	glNewList(model->data, GL_COMPILE);
	{
		for (int i = 0 ; i < numberOfPoints ; i += 3)
		{
			for (int j = 0 ; j < 3 ; j++)
			{
				token = strtok(NULL, "\n");
				
				sscanf(token, "%f %f %f %f %f", &x[j], &y[j], &z[j], &u[j], &v[j]);
			}
			
			glBegin(GL_TRIANGLES);
			{
				glTexCoord2f(u[0], v[0]); glVertex3f(x[0], y[0], z[0]);
				glTexCoord2f(u[1], v[1]); glVertex3f(x[1], y[1], z[1]);
				glTexCoord2f(u[2], v[2]); glVertex3f(x[2], y[2], z[2]);
			}
			glEnd();
		}
	}
	glEndList();
}

void createDisplayListFromProperties(char *filename, SimpleModel *model)
{
	List *list = loadProperties(filename, true);
	
	Vector bbMins, bbMaxs;
	Vector position, scale, size;
	float radius;
	GLColor color, rotation;
	String texture;
	
	model->data = glGenLists(1);
	
	//model->name = filename;
	
	glNewList(model->data, GL_COMPILE);
	{
		for (Properties *props = (Properties*)list->getFirstElement() ; props != NULL ; props = (Properties*)props->next)
		{
			glPushMatrix();
			{
				position = props->getVector("position");
				scale = props->getVector("scale");
				color = props->getColor("color");
				rotation = props->getColor("rotate"); // we're using a GLColor here because it can hold four values ;)
				
				if (rotation.color[0] != 1)
				{
					glRotatef(rotation.color[0], rotation.color[1], rotation.color[2], rotation.color[3]);
				}
				
				if ((scale.x != 0 ) && (scale.y != 0) && (scale.z != 0))
				{
					glScalef(scale.x, scale.y, scale.z);
				}
				
				glTranslatef(position.x, position.y, position.z);
				
				glColor4fv(color.color);
				
				if (props->name == "Sphere")
				{
					radius = props->getFloat("radius", 1);
					
					gluSphere(graphics->quadric, radius, 16, 16);
				}
				else if (props->name == "Box")
				{
					size = props->getVector("size");
					
					// note - This is done on purpose to ensure the values read are
					// in x, y, z rather than x, z, y
					//drawBox(position, size.x, size.z, size.y);
					drawBox(position, size.x, size.y, size.z);
				}
			}
			glPopMatrix();
		}
	}
	glEndList();
}

SimpleModel *loadSimpleModel(char *filename, int type)
{
	debug(("loadSimpleModel() - %s\n", filename));
	
	SimpleModel *simpleModel = new SimpleModel();
	
	switch (type)
	{
		case MDL_RAW:
			createDisplayListFromSimpleModel(filename, simpleModel);
			break;
			
		case MDL_PRIMITIVE:
			createDisplayListFromProperties(filename, simpleModel);
			break;
	}
	
	return simpleModel;
}
