/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "sentryGuns.h"

void sentryGunDie2()
{
	Vector v = self->position;
	v.x += Math::rrand(-3, 3);
	v.y += Math::rrand(-3, 3);
	
	addSmokeParticle(v);
	
	self->custom -= engine->getTimeDifference(TD_LOGIC);
	
	if (self->custom <= 0)
	{
		for (int i = 0 ; i < 25 ; i++)
		{
			throwExplosionDebris(self->position);
		}
		
		Math::addBit(&self->flags, EF_DEAD);
		
		for (int i = 0 ; i < 10 ; i++)
		{
			v = self->position;
			v.x += Math::rrand((int)self->boundingBox.mins.x, (int)self->boundingBox.maxs.x);
			v.y += Math::rrand((int)self->boundingBox.mins.y, (int)self->boundingBox.maxs.y);
			v.z += Math::rrand((int)self->boundingBox.mins.z, (int)self->boundingBox.maxs.z);
			doExplosion(v, 10, self);
		}
		
		audio->playSound(SND_LARGE_EXPLOSION, CH_ANY, camera->getSoundDistance(self->position));
		
		Unit *unit = (Unit*)self;
		
		if (unit->currentItem != NULL)
		{
			unit->currentItem->position = unit->position;
		
			unit->currentItem->setOwner(NULL);

			throwSpawnedItem(unit->currentItem);

			unit->currentItem = NULL;
		}
	}
}

void sentryGunDie1()
{
	Vector v = self->position;
	
	v.x += Math::rrand((int)self->boundingBox.mins.x, (int)self->boundingBox.maxs.x);
	v.y += Math::rrand((int)self->boundingBox.mins.y, (int)self->boundingBox.maxs.y);
	v.z += Math::rrand((int)self->boundingBox.mins.z, (int)self->boundingBox.maxs.z);
	doExplosion(v, 10, self);
	
	self->thinkTime = Math::rrand(15, 30);
	
	if (--self->custom == 0)
	{
		self->action = &sentryGunDie2;
		
		self->custom = 200;
	}
}

void sentryGunPostDie()
{
	self->health = 0;
	
	self->thinkTime = 0;
	
	self->action = sentryGunDie1;
	
	self->custom = Math::rrand(3, 5);
	
	Math::addBit(&self->flags, EF_DYING);
}

void sentryGunSearch()
{
	Unit *unit = (Unit*)self;
	
	if (unit->flags & EF_NOTARGET)
	{
		return;
	}
	
	if (unit->target != NULL)
	{
		faceTarget(unit);
	}
	
	if (enemySightTarget(unit, 650))
	{
		self->action = self->attack;
	}
	
	self->thinkTime = Math::rrand(25, 75);
}

void sentryGunDoNothing()
{
}

// ==================== Uzi sentry gun =================

void uziSentryGunAttack()
{
	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}

	self->thinkTime = 0;
	self->action = &uziSentryGunAttack;

	attackTarget(unit);

	if (--self->custom == 0)
	{
		self->action = &sentryGunSearch;
		self->thinkTime = Math::rrand(50, 100);
	}
}

void uziSentryGunPreAttack()
{
	self->custom = Math::rrand(10, 20);
	self->thinkTime = 0;
	self->action = &uziSentryGunAttack;
}

void uziSentryGunInit(Unit *unit)
{
	unit->walk = &sentryGunSearch;
	unit->attack = &uziSentryGunPreAttack;
	unit->die = &sentryGunPostDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}

// ==================== Plasma sentry gun =================

void plasmaSentryGunAttack()
{
	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}

	self->thinkTime = 0;
	self->action = &plasmaSentryGunAttack;

	attackTarget(unit);

	if (--self->custom == 0)
	{
		self->action = &sentryGunSearch;
		self->thinkTime = Math::rrand(50, 100);
	}
}

void plasmaSentryGunPreAttack()
{
	self->custom = Math::rrand(5, 20);
	self->thinkTime = 0;
	self->action = &plasmaSentryGunAttack;
}

void plasmaSentryGunInit(Unit *unit)
{
	unit->walk = &sentryGunSearch;
	unit->attack = &plasmaSentryGunPreAttack;
	unit->die = &sentryGunPostDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}

// ==================== Rocket sentry gun =================

void rocketSentryGunAttack()
{
	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}

	self->thinkTime = 0;
	self->action = &rocketSentryGunAttack;

	attackTarget(unit);

	if (--self->custom == 0)
	{
		self->action = &sentryGunSearch;
		self->thinkTime = Math::rrand(300, 500);
	}
}

void rocketSentryGunPreAttack()
{
	self->custom = Math::rrand(1, 2);
	self->thinkTime = 0;
	self->action = &rocketSentryGunAttack;
}

void rocketSentryGunInit(Unit *unit)
{
	unit->walk = &sentryGunSearch;
	unit->attack = &rocketSentryGunPreAttack;
	unit->die = &sentryGunPostDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}

// ==================== Mortor sentry gun =================

void mortorSentryGunAttack()
{
	Unit *unit = (Unit*)self;

	faceTarget(unit);

	if (unit->reload > 0)
	{
		return;
	}

	self->thinkTime = 0;
	self->action = &mortorSentryGunAttack;

	attackTarget(unit);

	if (--self->custom == 0)
	{
		self->action = &sentryGunSearch;
		self->thinkTime = Math::rrand(50, 100);
	}
}

void mortorSentryGunPreAttack()
{
	self->custom = Math::rrand(1, 3);
	self->thinkTime = 0;
	self->action = &mortorSentryGunAttack;
}

void mortorSentryGunInit(Unit *unit)
{
	unit->walk = &sentryGunSearch;
	unit->attack = &mortorSentryGunPreAttack;
	unit->die = &sentryGunPostDie;

	unit->thinkTime = Math::rrand(0, 50);

	unit->action = unit->walk;
}
